#include "Analysis.h"
#include "WaveHandler.h"
#include <math.h>
//Analysis.java
//Perform frequency analysis on digitized sound samples using autocorrelation
//
//(c) 2005 David Cuny
//http://tunafish.sourceforge.net
//Released under Qt Public License

// FIXME: silence minimum width should depend on sample rate
#define FREQ_SILENCE -1
#ifndef LOWEST_MIDI
#define LOWEST_MIDI 5
#endif
WaveAnalysis::WaveAnalysis( float* samples, int sampleslen, int sampleRate, int windowSize,
				WaveHandler* handler ) {
	// set the samples
	this->samples = samples;
	this->sampleslen = sampleslen;
	
	// save the sample rate
	this->sampleRate = sampleRate;
	
	// save the window size
	this->windowSize = windowSize;
	
	// marks width at start of wave
	this->waveWidth = new int[sampleslen];
	
	// holds the calculated frequency
	this->frequency = new float[sampleslen];
	
	// holds the frequency as a midi value
	this->midiValue = new double[sampleslen];		
	
	// holds the corrected frequency as a midi value
	this->correctedMidiValue = new double[sampleslen];

	for (int i=0;i<sampleslen;i++) 
	{
		waveWidth[i]=0;
		frequency[i]=0;
		midiValue[i]=0;
		correctedMidiValue[i]=0;
	}
	// find the maximum range of the wave
	this->findMaxAmplitude();
	
	// no restriction on scale
	this->scalePitch = (bool*)malloc(12*sizeof(bool));

	this->setScale("none");
	this->setScale("g major");

	this->maxAmplitude = 0;
	this->zero = 0;
	this->zeroCutoff = 0;
	this->silenceMinWidth = 200;
	this->segments = new vector<WaveSegment*>();
	this->needsResampling = true;

	this->wavehandler=handler;
}

WaveAnalysis::~WaveAnalysis()
{
	free(this->scalePitch);
	delete this->waveWidth;
	delete this->frequency;
	delete this->midiValue;
	delete this->correctedMidiValue;
	delete this->scalePitch;
	delete this->segments;
}
	
void WaveAnalysis::setScale( const char* keyName ) 
{
	// scale should always be given in lowercase.		
	if (0==strcmp(keyName,"chromatic")) {
		// no restrictions
		for (int i = 0; i < 12; i++) {
			this->scalePitch[i] = true;
		}
		return;
	}
	
	// restrict all pitches
	for (int i = 0; i < 12; i++) {
		this->scalePitch[i] = false;
	}
	
	if (0==strcmp(keyName,"c major")) {
		// c d e f g a b
		this->scalePitch[WaveAnalysis::C_NOTE] 		= true;			
		this->scalePitch[WaveAnalysis::D_NOTE] 		= true;			
		this->scalePitch[WaveAnalysis::E_NOTE] 		= true;			
		this->scalePitch[WaveAnalysis::F_NOTE] 		= true;			
		this->scalePitch[WaveAnalysis::G_NOTE] 		= true;			
		this->scalePitch[WaveAnalysis::A_NOTE] 		= true;			
		this->scalePitch[WaveAnalysis::B_NOTE] 		= true;			
		return;
		
	} else if (0==strcmp(keyName,"c# major")) {
		// c# d# e# f# g# a# b#
		this->scalePitch[WaveAnalysis::C_SHARP_NOTE] 	= true;			
		this->scalePitch[WaveAnalysis::D_SHARP_NOTE] 	= true;			
		this->scalePitch[WaveAnalysis::E_SHARP_NOTE]	= true;			
		this->scalePitch[WaveAnalysis::F_SHARP_NOTE] 	= true;			
		this->scalePitch[WaveAnalysis::G_SHARP_NOTE] 	= true;			
		this->scalePitch[WaveAnalysis::A_SHARP_NOTE] 	= true;			
		this->scalePitch[WaveAnalysis::B_SHARP_NOTE]	= true;			
		return;
			
	} else if (0==strcmp(keyName,"d major")) {
		// d e f# g a b c#
		this->scalePitch[WaveAnalysis::D_NOTE] 		= true;			
		this->scalePitch[WaveAnalysis::E_NOTE] 		= true;			
		this->scalePitch[WaveAnalysis::F_SHARP_NOTE]	= true;			
		this->scalePitch[WaveAnalysis::G_NOTE] 		= true;			
		this->scalePitch[WaveAnalysis::A_NOTE] 		= true;			
		this->scalePitch[WaveAnalysis::B_NOTE] 		= true;			
		this->scalePitch[WaveAnalysis::C_SHARP_NOTE] 	= true;			
		return;

	} else if (0==strcmp(keyName,"d# major")) {
		// d# e# fx=g g# a# b# c#
		this->scalePitch[WaveAnalysis::D_SHARP_NOTE]	= true;			
		this->scalePitch[WaveAnalysis::E_SHARP_NOTE] 	= true;			
		this->scalePitch[WaveAnalysis::G_NOTE]		= true;			
		this->scalePitch[WaveAnalysis::G_SHARP_NOTE]	= true;			
		this->scalePitch[WaveAnalysis::A_SHARP_NOTE]	= true;			
		this->scalePitch[WaveAnalysis::B_SHARP_NOTE] 	= true;			
		this->scalePitch[WaveAnalysis::C_SHARP_NOTE] 	= true;			
		return;

	} else if (0==strcmp(keyName,"e major")) {
		// e f# g# a b c# d#
		this->scalePitch[WaveAnalysis::E_NOTE] 		= true;			
		this->scalePitch[WaveAnalysis::F_SHARP_NOTE] 	= true;			
		this->scalePitch[WaveAnalysis::G_SHARP_NOTE]	= true;			
		this->scalePitch[WaveAnalysis::A_NOTE] 		= true;			
		this->scalePitch[WaveAnalysis::B_NOTE] 		= true;			
		this->scalePitch[WaveAnalysis::C_SHARP_NOTE] 	= true;			
		this->scalePitch[WaveAnalysis::D_SHARP_NOTE] 	= true;			
		return;

	} else if (0==strcmp(keyName,"f major")) {
		// f g a bb c d e
		this->scalePitch[WaveAnalysis::F_NOTE] 		= true;			
		this->scalePitch[WaveAnalysis::G_NOTE] 		= true;			
		this->scalePitch[WaveAnalysis::A_NOTE]		= true;			
		this->scalePitch[WaveAnalysis::B_FLAT_NOTE] 	= true;			
		this->scalePitch[WaveAnalysis::C_NOTE] 		= true;			
		this->scalePitch[WaveAnalysis::D_NOTE] 		= true;			
		this->scalePitch[WaveAnalysis::E_NOTE] 		= true;			
		return;

	} else if (0==strcmp(keyName,"f# major")) {
		// f# g# a# b c# d# e#
		this->scalePitch[WaveAnalysis::F_SHARP_NOTE] 	= true;			
		this->scalePitch[WaveAnalysis::G_SHARP_NOTE] 	= true;			
		this->scalePitch[WaveAnalysis::A_SHARP_NOTE]	= true;			
		this->scalePitch[WaveAnalysis::B_NOTE] 		= true;			
		this->scalePitch[WaveAnalysis::C_SHARP_NOTE] 	= true;			
		this->scalePitch[WaveAnalysis::D_SHARP_NOTE] 	= true;			
		this->scalePitch[WaveAnalysis::E_SHARP_NOTE] 	= true;			
		return;

	} else if (0==strcmp(keyName,"g major")) {
		// g a b c d e f# 
		this->scalePitch[WaveAnalysis::G_NOTE] 		= true;			
		this->scalePitch[WaveAnalysis::A_NOTE] 		= true;			
		this->scalePitch[WaveAnalysis::B_NOTE]		= true;			
		this->scalePitch[WaveAnalysis::C_NOTE] 		= true;			
		this->scalePitch[WaveAnalysis::D_NOTE] 		= true;			
		this->scalePitch[WaveAnalysis::E_NOTE] 		= true;			
		this->scalePitch[WaveAnalysis::F_SHARP_NOTE] 	= true;			
		return;
		
	} else if (0==strcmp(keyName,"g# major")) {
		// g# a# b# c# d# e# fx=g
		this->scalePitch[WaveAnalysis::G_SHARP_NOTE] 	= true;			
		this->scalePitch[WaveAnalysis::A_SHARP_NOTE] 	= true;			
		this->scalePitch[WaveAnalysis::B_SHARP_NOTE]	= true;			
		this->scalePitch[WaveAnalysis::C_SHARP_NOTE] 	= true;			
		this->scalePitch[WaveAnalysis::D_SHARP_NOTE] 	= true;			
		this->scalePitch[WaveAnalysis::E_SHARP_NOTE] 	= true;			
		this->scalePitch[WaveAnalysis::G_NOTE] 		= true;			
		return;

	} else if (0==strcmp(keyName,"a major")) {
		// a b c# d e f# g#			
		this->scalePitch[WaveAnalysis::A_NOTE] 		= true;			
		this->scalePitch[WaveAnalysis::B_NOTE]		= true;			
		this->scalePitch[WaveAnalysis::C_SHARP_NOTE]	= true;			
		this->scalePitch[WaveAnalysis::D_NOTE] 		= true;			
		this->scalePitch[WaveAnalysis::E_NOTE] 		= true;			
		this->scalePitch[WaveAnalysis::F_SHARP_NOTE] 	= true;
		this->scalePitch[WaveAnalysis::G_SHARP_NOTE]	= true;
		return;

	} else if (0==strcmp(keyName,"a# major")) {
		// a# b# cx=d d# e# fx=g gx=a			
		this->scalePitch[WaveAnalysis::A_SHARP_NOTE] 	= true;			
		this->scalePitch[WaveAnalysis::B_SHARP_NOTE]	= true;			
		this->scalePitch[WaveAnalysis::D_NOTE]		= true;		
		this->scalePitch[WaveAnalysis::D_SHARP_NOTE] 	= true;			
		this->scalePitch[WaveAnalysis::E_SHARP_NOTE] 	= true;			
		this->scalePitch[WaveAnalysis::G_NOTE] 		= true;
		this->scalePitch[WaveAnalysis::A_NOTE]		= true;
		return;

	} else if (0==strcmp(keyName,"b major")) {
		// b c# d# e f# g# a			
		this->scalePitch[WaveAnalysis::B_NOTE] 		= true;			
		this->scalePitch[WaveAnalysis::C_SHARP_NOTE]	= true;			
		this->scalePitch[WaveAnalysis::D_SHARP_NOTE]	= true;		
		this->scalePitch[WaveAnalysis::E_NOTE] 		= true;			
		this->scalePitch[WaveAnalysis::F_SHARP_NOTE] 	= true;			
		this->scalePitch[WaveAnalysis::G_SHARP_NOTE] 	= true;
		this->scalePitch[WaveAnalysis::A_NOTE]		= true;
		return;
	}
		
}
	
void WaveAnalysis::findMaxAmplitude() {

	this->zero = 0;//samples[0];
	int maxSample = this->zero;
	int minSample = this->zero;
	
	for (int i = 0; i < this->sampleslen; i++) {
		float value = samples[i];			
					
		// maximum value?
		if (value > maxSample) {
			maxSample = value;
		}
		
		if (value < minSample) {
			minSample = value;
		}
	}
	
	// save max value

	this->maxAmplitude = max(maxSample, -minSample );

	// set the zero cutoff as 10% of the maximum amplitude
	this->zeroCutoff = (int)((double)this->maxAmplitude * .10);		
	
}

double WaveAnalysis::closestMatch( double midiPitch) {
	// return closest matching midi scale note
	// called by: calculateCorrectPitch

	// FIXME: shouldn't happen, but it does.
	if (midiPitch < LOWEST_MIDI) {
		return -1;
	}
	
	// find the exact midi pitch
	double centeredMidi = floor(midiPitch + (double)0.5);
	
	// find the note number (0..12)
	int scaleNumber = ((int)(centeredMidi)) % 12;
	
	// part of the allowed pitches in the scale?
	if (this->scalePitch[scaleNumber]) {
		// return as good match
		return centeredMidi;
	}
	
	// FIXME: this assumes there's a good pitch a half step above or below
	// FIXME: this should go to the *CLOSEST* match
	
	// get scale number of pitch above and below
	double toAbove =  midiPitch - (centeredMidi + 1) ;
	if (toAbove<0) { toAbove=-toAbove; }
	double toBelow =  midiPitch - (centeredMidi - 1) ;
	if (toBelow<0) { toBelow=-toBelow; }
	
	// return the closest pitch
	if (toAbove < toBelow) {
		return centeredMidi + 1;
	}
	return centeredMidi - 1;		
}
	
	
void WaveAnalysis::markFrequency( int start, int end ) {
	// calculate width
	int width = end - start;
	
	// flag start of wave
	this->waveWidth[start] = width;
	
	// calculate the frequency
	float freq = sampleRate / (float)width; 
	
	// iterate over the range
	for (int i = start; i <= end; i++ ) {
		// set the frequency
		frequency[i]= freq;
	}
}
	
		
bool WaveAnalysis::near( float error, float a, float b ) {
	// return true if a is near b
	return !(this->myabs((double)a-(double)b) > error);
}



/**
 * This compares the current pitch (represented as a MIDI number) to the target
 * pitch (supplied by the method <b>closestMatch</b>. If they differ, the pitch is
 * nudged in the direction of the target pitch by the <b>correctionAmount</b>. This
 * is done using MIDI pitches instead of actual frequencies, so that the response to
 * the pitch correction occurs in linear space. Otherwise, the pitch correction would
 * be faster for higher pitches, since the distance between pitches decreases as the
 * pitch frequency increases.
 * @param correctionAmount
 */
double WaveAnalysis::myabs(double x)
{
	if (x>=0) return x;
	return -x;
}

void WaveAnalysis::calculateCorrectPitch() { 
	// called by UI after wave segment drag and after scale change.

	// flag need to recalculate wave
	if (this==NULL)
	{
		return;
	}
	this->needsResampling = true;
	
	// create an array to hold the segment midi values
	double* segmentMidi = new double[this->sampleslen];
	for (int i = 0; i < this->sampleslen; i++) {
		segmentMidi[i] = -1;
	}
	
	// iterate over the segments
	for (int i = 0; i < this->segments->size(); i++) {
		// get a wave segment
		vector<WaveSegment*>* thisvec=this->segments;
		WaveSegment* segment = (*thisvec)[i];
		
		// copy the midi value over the range of the segment
		for (int j = segment->startSample; j <= segment->endSample; j++ ) {
			segmentMidi[j] = segment->midiNumber;
		}
		
	}
	
	
	// don't change the starting frequency
	double currentMidi = midiValue[0];
	correctedMidiValue[0] = currentMidi;

	// target is the closest matching pitch
	double targetMidi = -1;

	// iterate through the signal
	for (int i = 1; i < sampleslen; i++) 
	{
		// look at the midi value
		if (this->midiValue[i] < LOWEST_MIDI) {
			// noise or silence
			currentMidi = -1;
			targetMidi = -1;
		} 
		else if (midiValue[i-1] < LOWEST_MIDI) {
			// start of a new signal, begin at this frequency with no correction
			// currentMidi = midiValue[i];
			
			// FIXME: this starts the pitch at the correct midi value, to avoid scooping
			// there should probably be some logic that checks for the difference to 
			// decide if some scooping should be left in
			currentMidi = closestMatch(segmentMidi[i]);
		}
		else if (this->myabs(segmentMidi[i-1]-segmentMidi[i]) > 1) {
			// skip up to target
			currentMidi = closestMatch(segmentMidi[i]);
 		} else {
			// continuing signal, add delta to get change
			currentMidi += midiValue[i] - midiValue[i-1];
		
			// find target based on segment's midi value
			targetMidi = closestMatch(segmentMidi[i]);
		
			if (currentMidi == targetMidi) {
				// already matches, do nothing
	
	
			} else if (currentMidi < targetMidi) {
				// less than desired frequency, so add correction amount
				currentMidi += (wavehandler->CORRECTION_RATE);
			
				// overshot the target frequency?
				if (currentMidi > targetMidi) {
					// set the current frequency to the target 	frequency
					currentMidi = targetMidi;
				}
					
			} else {
				// greater than the desired amount, so subtract correction 	amount
				currentMidi -= (wavehandler->CORRECTION_RATE);
			
				// undershoot the target?
				if (currentMidi < targetMidi) {
					// set the current amount to the target amount
					currentMidi = targetMidi;
				}

			}	
			// save the corrected pitch
			correctedMidiValue[i] = currentMidi;
		}	
	}		
	delete segmentMidi;
}

	
void WaveAnalysis::findSilence() {
	
	// set all frequency to zero
	for (int i = 0; i < this->sampleslen; i++) {
		this->frequency[i] = 0;
	}
	
	
	// iterate over the samples
	int startAt = 0;
	for (int at = 0; at < this->sampleslen; at++) {
		
		// get a sample
		float sample = this->samples[at];
		if (sample<0) sample=0-sample;

		// found a sample that's not silence?
		if ((sample > this->zeroCutoff) || (at == this->sampleslen-1)) {
			// was there a starting point?
			if (startAt != -1) {
				// enough silence to mark?
				int wide = at - startAt;
				if (wide > this->silenceMinWidth) {						
					for (int i = startAt; i < at; i++) {
						// mark as silence
						frequency[i] = FREQ_SILENCE;
					}
				}
				
				// reset start at position
				startAt = -1;
			} 			
		// need to reset startAt at beginning of silence?
		} else if (startAt == -1) {
			// new start of silence
			startAt = at;
		}
	}
}

	
int WaveAnalysis::autocorrelate( int at, int localWindowSize ) {
	// called by findFrequencies
	// past range?
	if (at+localWindowSize > this->sampleslen) {
		return -1;
	}
	
	localWindowSize = localWindowSize / 2;
	double maximum = -1;
	int maximumAt = -1;		

	// skip if there is any silence in the window
	for (int i = at; i < at+localWindowSize; i++) {
		if (frequency[i] == FREQ_SILENCE) {
			return -1;
		}
	}
	
	
	// range to examine
	int start = at + 20;
	int end = at + localWindowSize;
	
	// slide window
	for (int i = start; i < end; i++ ) {

		// calculate the difference
		double sum = 0;
		for (int j=0; j < localWindowSize; j++) {
			sum += samples[at+j] * samples[i+j];				
		}
					
		// maximum value so far?
		if (sum > maximum) {				
			// save value
			maximum = sum;
			maximumAt = i;
		}
	}
	
	return maximumAt;
	
}
		

	
void WaveAnalysis::findFrequencies() {
		
	// reset frequencies and mark sections as silence
	this->findSilence();
	
	int priorWidth = -1;
	
	// iterate until can't correlate any more
	for (int i = 0; i < this->sampleslen; i++ ) 
	{		
		// not silence?
		if (frequency[i] == FREQ_SILENCE) continue;
		
		// what width to pass to autocorrelation?
		int wide = priorWidth;
		if (priorWidth < 30) {
			// prior width too small, use default window size		
			wide = this->windowSize;
		} else {
			// use 3*priorWidth
			wide = (int)((float)priorWidth * (float)3);
		}
		
		// find match using autocorrelation
		int match = autocorrelate(i, wide);
		if (match==-1)
		{
			// there was silence in the window.
			continue;
		}

		// mark the frequency as a match (where i,match indicate 2 points in wave)
		markFrequency( i, match );
	
		// save as prior width
		priorWidth = match - i;
	
		// move ahead
		i = match - 1;
	}
	
	// mark all frequencies of zero as silence
	// MRJB: this does not make sense. (shouldn't this be this->frequency[i]==0?)
	for (int i = 0; i < this->sampleslen; i++) {
		//if (this->samples[i] == 0) {
//			this->samples[i] = -1;
		//}
		if (this->frequency[i]==0) {
			this->frequency[i]=-1;
		}
	}

	// convert frequency to pitch numbers
	for (int i = 0; i < this->sampleslen; i++) {
		if (this->frequency[i] > 10) {
			this->midiValue[i] = frequencyToMidi( this->frequency[i] );
	
		} else {
			this->midiValue[i] = -1;
		}			
	}
	
}
	
	
void WaveAnalysis::findSegments() {
	// clear the segments
	this->segments = new vector<WaveSegment*>();
	
	// start segment not set
	int startAt = -1;
	double midi = 0;
	double midiSum = 0;
	
	// iterate through the 
	for (int i = 0; i < this->sampleslen; i++) {
		// found silence?
		double myabs=this->myabs(midiValue[i]-midi);
		if (this->frequency[i] == -1 || 
			(startAt != -1 && (myabs > 2.0) )) {
			 // got a start segment?
			if (startAt != -1) {
				// how wide is the segment?
				// FIXME: need real value at some point, not 1000
				if ( i - startAt > 1000) {
					int width = i - startAt;
					int averagedMidi = floor((midiSum / width) + 0.5);
					WaveSegment* segment = new WaveSegment(startAt, i-1, averagedMidi, this->samples,this->sampleslen);
					this->segments->push_back(segment);
				}
			}
			// clear the start
			startAt = -1;
		} else {
			// need to set start position?
			if (startAt == -1) {
				// set start
				startAt = i;
				// FIXME: need to find median value or something...
				midi = this->midiValue[i];
				midiSum = this->midiValue[i];
			} else {
				// accumulate midi value
				midiSum += this->midiValue[i];
			}
		}
	}
	
}
	
double WaveAnalysis::frequencyToMidi( float f ) {
	return ((12.0 * log(f/220.0)) / log(2.0)) + (double)57.01; 
}

double WaveAnalysis::midiToFrequency( double midi ) {
	return (double)440.0 * pow(2, (midi-69)/12 );
}
	

