import java.util.Vector;

//Analysis.java
//Perform frequency analysis on digitized sound samples using autocorrelation
//
//(c) 2005 David Cuny
//http://tunafish.sourceforge.net
//Released under Qt Public License

// FIXME: silence minimum width should depend on sample rate

public class Analysis {
			
	// lowest possible midi frequency
	final static int LOWEST_MIDI = 5;
	
	float sampleRate;
	int windowSize;
	int[] samples;
	float[] frequency;
	double[] midiValue;
	double[] correctedMidiValue;
	int[] waveWidth;
	int maxAmplitude = 0;
	int zero = 0;
	int zeroCutoff = 0;
	int silenceMinWidth = 200;
	Vector segments = new Vector();
	boolean needsResampling = true;
	
	// restricted pitches and pitch names
	boolean scalePitch[] = new boolean[12];
	final static int C_NOTE = 0;
	
	final static int C_SHARP_NOTE = 1;
	final static int D_FLAT_NOTE = 1;

	final static int D_NOTE = 2;
	
	final static int D_SHARP_NOTE = 3;
	final static int E_FLAT_NOTE = 3;
	
	final static int E_NOTE = 4;

	final static int E_SHARP_NOTE = 5;
	final static int F_FLAT_NOTE = 4;
	
	final static int F_NOTE = 5;
	
	final static int F_SHARP_NOTE = 6;
	final static int G_FLAT_NOTE = 6;
	
	final static int G_NOTE = 7;
	
	final static int G_SHARP_NOTE = 8;
	final static int A__FLAT_NOTE = 9;
	
	final static int A_NOTE = 9;
	
	final static int A_SHARP_NOTE = 10;
	final static int B_FLAT_NOTE = 10;
	
	final static int B_NOTE = 11;
	
	final static int B_SHARP_NOTE = 0;
	final static int C_FLAT_NOTE = 11;
	
	
	Analysis( int[] samples, float sampleRate, int windowSize ) {
		// set the samples
		this.samples = samples;
		
		// save the sample rate
		this.sampleRate = sampleRate;
		
		// save the window size
		this.windowSize = windowSize;
		
		// marks width at start of wave
		this.waveWidth = new int[samples.length];
		
		// holds the calculated frequency
		this.frequency = new float[samples.length];
		
		// holds the frequency as a midi value
		this.midiValue = new double[samples.length];		
		
		// holds the corrected frequency as a midi value
		this.correctedMidiValue = new double[samples.length];

		// find the maximum range of the wave
		this.findMaxAmplitude();
		
		// no restriction on scale
		this.setScale("none");
		this.setScale("g major");
		
	}

	
	public void setScale( String keyName ) {
		
		if (keyName.equalsIgnoreCase("chromatic")) {
			// no restrictions
			for (int i = 0; i < 12; i++) {
				scalePitch[i] = true;
			}
			return;
		}
		
		// restrict all pitches
		for (int i = 0; i < 12; i++) {
			scalePitch[i] = false;
		}
		
		if (keyName.equalsIgnoreCase("c major")) {
			// c d e f g a b
			this.scalePitch[C_NOTE] 		= true;			
			this.scalePitch[D_NOTE] 		= true;			
			this.scalePitch[E_NOTE] 		= true;			
			this.scalePitch[F_NOTE] 		= true;			
			this.scalePitch[G_NOTE] 		= true;			
			this.scalePitch[A_NOTE] 		= true;			
			this.scalePitch[B_NOTE] 		= true;			
			return;
			
		} else if (keyName.equalsIgnoreCase("c# major")) {
			// c# d# e# f# g# a# b#
			this.scalePitch[C_SHARP_NOTE] 	= true;			
			this.scalePitch[D_SHARP_NOTE] 	= true;			
			this.scalePitch[E_SHARP_NOTE]	= true;			
			this.scalePitch[F_SHARP_NOTE] 	= true;			
			this.scalePitch[G_SHARP_NOTE] 	= true;			
			this.scalePitch[A_SHARP_NOTE] 	= true;			
			this.scalePitch[B_SHARP_NOTE]	= true;			
			return;
				
		} else if (keyName.equalsIgnoreCase("d major")) {
			// d e f# g a b c#
			this.scalePitch[D_NOTE] 		= true;			
			this.scalePitch[E_NOTE] 		= true;			
			this.scalePitch[F_SHARP_NOTE]	= true;			
			this.scalePitch[G_NOTE] 		= true;			
			this.scalePitch[A_NOTE] 		= true;			
			this.scalePitch[B_NOTE] 		= true;			
			this.scalePitch[C_SHARP_NOTE] 	= true;			
			return;

		} else if (keyName.equalsIgnoreCase("d# major")) {
			// d# e# fx=g g# a# b# c#
			this.scalePitch[D_SHARP_NOTE]	= true;			
			this.scalePitch[E_SHARP_NOTE] 	= true;			
			this.scalePitch[G_NOTE]			= true;			
			this.scalePitch[G_SHARP_NOTE]	= true;			
			this.scalePitch[A_SHARP_NOTE]	= true;			
			this.scalePitch[B_SHARP_NOTE] 	= true;			
			this.scalePitch[C_SHARP_NOTE] 	= true;			
			return;

		} else if (keyName.equalsIgnoreCase("e major")) {
			// e f# g# a b c# d#
			this.scalePitch[E_NOTE] 		= true;			
			this.scalePitch[F_SHARP_NOTE] 	= true;			
			this.scalePitch[G_SHARP_NOTE]	= true;			
			this.scalePitch[A_NOTE] 		= true;			
			this.scalePitch[B_NOTE] 		= true;			
			this.scalePitch[C_SHARP_NOTE] 	= true;			
			this.scalePitch[D_SHARP_NOTE] 	= true;			
			return;

		} else if (keyName.equalsIgnoreCase("f major")) {
			// f g a bb c d e
			this.scalePitch[F_NOTE] 		= true;			
			this.scalePitch[G_NOTE] 		= true;			
			this.scalePitch[A_NOTE]			= true;			
			this.scalePitch[B_FLAT_NOTE] 	= true;			
			this.scalePitch[C_NOTE] 		= true;			
			this.scalePitch[D_NOTE] 		= true;			
			this.scalePitch[E_NOTE] 		= true;			
			return;

		} else if (keyName.equalsIgnoreCase("f# major")) {
			// f# g# a# b c# d# e#
			this.scalePitch[F_SHARP_NOTE] 	= true;			
			this.scalePitch[G_SHARP_NOTE] 	= true;			
			this.scalePitch[A_SHARP_NOTE]	= true;			
			this.scalePitch[B_NOTE] 		= true;			
			this.scalePitch[C_SHARP_NOTE] 	= true;			
			this.scalePitch[D_SHARP_NOTE] 	= true;			
			this.scalePitch[E_SHARP_NOTE] 	= true;			
			return;

		} else if (keyName.equalsIgnoreCase("g major")) {
			// g a b c d e f# 
			this.scalePitch[G_NOTE] 		= true;			
			this.scalePitch[A_NOTE] 		= true;			
			this.scalePitch[B_NOTE]			= true;			
			this.scalePitch[C_NOTE] 		= true;			
			this.scalePitch[D_NOTE] 		= true;			
			this.scalePitch[E_NOTE] 		= true;			
			this.scalePitch[F_SHARP_NOTE] 	= true;			
			return;
			
		} else if (keyName.equalsIgnoreCase("g# major")) {
			// g# a# b# c# d# e# fx=g
			this.scalePitch[G_SHARP_NOTE] 	= true;			
			this.scalePitch[A_SHARP_NOTE] 	= true;			
			this.scalePitch[B_SHARP_NOTE]	= true;			
			this.scalePitch[C_SHARP_NOTE] 	= true;			
			this.scalePitch[D_SHARP_NOTE] 	= true;			
			this.scalePitch[E_SHARP_NOTE] 	= true;			
			this.scalePitch[G_NOTE] 		= true;			
			return;

		} else if (keyName.equalsIgnoreCase("a major")) {
			// a b c# d e f# g#			
			this.scalePitch[A_NOTE] 		= true;			
			this.scalePitch[B_NOTE]			= true;			
			this.scalePitch[C_SHARP_NOTE]	= true;			
			this.scalePitch[D_NOTE] 		= true;			
			this.scalePitch[E_NOTE] 		= true;			
			this.scalePitch[F_SHARP_NOTE] 	= true;
			this.scalePitch[G_SHARP_NOTE]	= true;
			return;

		} else if (keyName.equalsIgnoreCase("a# major")) {
			// a# b# cx=d d# e# fx=g gx=a			
			this.scalePitch[A_SHARP_NOTE] 	= true;			
			this.scalePitch[B_SHARP_NOTE]	= true;			
			this.scalePitch[D_NOTE]			= true;		
			this.scalePitch[D_SHARP_NOTE] 	= true;			
			this.scalePitch[E_SHARP_NOTE] 	= true;			
			this.scalePitch[G_NOTE] 		= true;
			this.scalePitch[A_NOTE]			= true;
			return;

		} else if (keyName.equalsIgnoreCase("b major")) {
			// b c# d# e f# g# a			
			this.scalePitch[B_NOTE] 		= true;			
			this.scalePitch[C_SHARP_NOTE]	= true;			
			this.scalePitch[D_SHARP_NOTE]	= true;		
			this.scalePitch[E_NOTE] 		= true;			
			this.scalePitch[F_SHARP_NOTE] 	= true;			
			this.scalePitch[G_SHARP_NOTE] 	= true;
			this.scalePitch[A_NOTE]			= true;
			return;
		}
			
	}
	
	public void findMaxAmplitude() {

		this.zero = samples[0];
		int maxSample = zero;
		int minSample = zero;
		
		for (int i = 0; i < this.samples.length; i++) {
			int value = 0; 
			value = samples[i] - zero;			
						
			// maximum value?
			if (value > maxSample) {
				maxSample = value;
			}
			
			if (-value < minSample) {
				minSample = value;
			}
		}
		
		// save max value
		this.maxAmplitude = Math.max(maxSample, -minSample );

		// set the zero cutoff as 10% of the maximum amplitude
		this.zeroCutoff = (int)((double)this.maxAmplitude * .10);		
		
	}
	
	public double closestMatch( double midiPitch) {
		// return closest matching midi scale note
		
		// FIXME: shouldn't happen, but it does.
		if (midiPitch < LOWEST_MIDI) {
			return -1;
		}
		
		// find the exact midi pitch
		int centeredMidi = (int)(midiPitch + 0.5d);
		
		// find the note number (0..12)
		int scaleNumber = centeredMidi % 12;
		
		// part of the allowed pitches in the scale?
		if (this.scalePitch[scaleNumber]) {
			// return as good match
			return (double)centeredMidi;
		}
		
		// FIXME: this assumes there's a good pitch a half step above or below
		// FIXME: this should go to the *CLOSEST* match
		
		// get scale number of pitch above and below
		double toAbove = Math.abs( midiPitch - ((double)centeredMidi + 1) );
		double toBelow = Math.abs( midiPitch - ((double)centeredMidi - 1) );
		
		// return the closest pitch
		if (toAbove < toBelow) {
			return centeredMidi + 1;
		} else {
			return centeredMidi - 1;
		}
		
		
	}
	
	
	public void markFrequency( int start, int end ) {
		// calculate width
		int width = end - start;
		
		// flag start of wave
		this.waveWidth[start] = width;
		
		// calculate the frequency
		float freq = sampleRate / (float)width; 
		
		// iterate over the range
		for (int i = start; i <= end; i++ ) {
			// set the frequency
			frequency[i]= freq;
		}
	}
	
		
	public static boolean near( int error, int a, int b ) {
		// return true if a is near b
		return !(Math.abs(a-b) > error);
	}

	

	/**
	 * This compares the current pitch (represented as a MIDI number) to the target
	 * pitch (supplied by the method <b>closestMatch</b>. If they differ, the pitch is
	 * nudged in the direction of the target pitch by the <b>correctionAmount</b>. This
	 * is done using MIDI pitches instead of actual frequencies, so that the response to
	 * the pitch correction occurs in linear space. Otherwise, the pitch correction would
	 * be faster for higher pitches, since the distance between pitches decreases as the
	 * pitch frequency increases.
	 * @param correctionAmount
	 */
	public void calculateCorrectPitch() { 
		
		// flag need to recalculate wave
		this.needsResampling = true;
		
		// create an array to hold the segment midi values
		double[] segmentMidi = new double[this.samples.length];
		for (int i = 0; i < this.samples.length; i++) {
			segmentMidi[i] = -1;
		}
		
		// iterate over the segments
		for (int i = 0; i < this.segments.size(); i++) {
			// get a wave segment
			WaveSegment segment = (WaveSegment)segments.get(i);
			
			// copy the midi value over the range of the segment
			for (int j = segment.startSample; j <= segment.endSample; j++ ) {
				segmentMidi[j] = segment.midiNumber;
			}
			
		}
		
		
		// don't change the starting frequency
		double currentMidi = midiValue[0];
		correctedMidiValue[0] = currentMidi;

		// target is the closest matching pitch
		double targetMidi = -1;

		
		// iterate through the signal
		for (int i = 1; i < samples.length; i++) {
			
			// look at the midi value
			if (this.midiValue[i] < LOWEST_MIDI) {
				// noise or silence
				currentMidi = -1;
				targetMidi = -1;
				
			} else if (midiValue[i-1] < LOWEST_MIDI) {
				// start of a new signal, begin at this frequency with no correction
				// currentMidi = midiValue[i];
				
				// FIXME: this starts the pitch at the correct midi value, to avoid scooping
				// there should probably be some logic that checks for the difference to 
				// decide if some scooping should be left in
				currentMidi = closestMatch(segmentMidi[i]);
				

			} else if (Math.abs(segmentMidi[i-1]-segmentMidi[i]) > 1) {
				// skip up to targe
				currentMidi = closestMatch(segmentMidi[i]);

			} else {
				// continuing signal, add delta to get change
				currentMidi += midiValue[i] - midiValue[i-1];
				
				// find target based on segment's midi value
				targetMidi = closestMatch(segmentMidi[i]);
				
				if (currentMidi == targetMidi) {
					// already matches, do nothing
					
				} else if (currentMidi < targetMidi) {
					// less than desired frequency, so add correction amount
					currentMidi += WaveHandler.CORRECTION_RATE;
					
					// overshot the target frequency?
					if (currentMidi > targetMidi) {
						// set the current frequency to the target frequency
						currentMidi = targetMidi;
					}
					
				} else {
					// greater than the desired amount, so subtract correction amount
					currentMidi -= WaveHandler.CORRECTION_RATE;
					
					// undershoot the target?
					if (currentMidi < targetMidi) {
						// set the current amount to the target amount
						currentMidi = targetMidi;
					}

				}
			
				// save the corrected pitch
				correctedMidiValue[i] = currentMidi;				
			}
		}		
	}

	
	public void findSilence() {
		
		// set all frequency to zero
		for (int i = 0; i < this.samples.length; i++) {
			this.frequency[i] = 0;
		}
		
		
		// iterate over the samples
		int startAt = 0;
		for (int at = 0; at < this.samples.length; at++) {
			
			// get a sample
			int sample = Math.abs(samples[at]-this.zero);
						
			// found a sample that's not silence?
			if ((sample > this.zeroCutoff) || (at == this.samples.length-1)) {
				// was there a starting point?
				if (startAt != -1) {
					// enough silence to mark?
					int wide = at - startAt;
					if (wide > this.silenceMinWidth) {						
						for (int i = startAt; i < at; i++) {
							// mark as silence
							frequency[i] = -1;
						}
					}
					
					// reset start at position
					startAt = -1;

				}
				
			// need to reset startAt at beginning of silence?
			} else if (startAt == -1) {
				// new start of silence
				startAt = at;
			}
		}
	}

	
	public int autocorrelate( int at, int localWindowSize ) {
		
		// past range?
		if (at+localWindowSize > this.samples.length) {
			return -1;
		}
		
		localWindowSize = localWindowSize / 2;
		double maximum = -1;
		int maximumAt = -1;		

		// skip if there is any silence in the window
		for (int i = at; i < at+localWindowSize; i++) {
			if (frequency[i] == -1) {
				return -1;
			}
		}
		
		
		// range to examine
		int start = at + 20;
		int end = at + localWindowSize;
		
		// slide window
		for (int i = start; i < end; i++ ) {

			// calculate the difference
			double sum = 0;
			for (int j=0; j < localWindowSize; j++) {
				sum += samples[at+j] * samples[i+j];				
			}
						
			// maximum value so far?
			if (sum > maximum) {				
				// save value
				maximum = sum;
				maximumAt = i;
			}
		}
		
		return maximumAt;
		
	}
		

	
	public void findFrequencies() {
		
		// reset frequencies and mark sections as silence
		this.findSilence();
		
		int priorWidth = -1;
		
		// iterate until can't correlate any more
		for (int i = 0; i < this.samples.length; i++ ) {
			
			// not silence?
			if (frequency[i] != -1) { 			
			
				// what width to pass to autocorrelation?
				int wide = priorWidth;
				if (priorWidth < 30) {
					// prior width too small, use default window size		
					wide = this.windowSize;
				} else {
					// use 3*priowWidth
					wide = (int)((float)priorWidth * 3f);
				}
				
				// find match using autocorrelation
				int match = autocorrelate(i, wide);
				if (match != -1) {
					// mark the frequency as a match
					markFrequency( i, match );
				
					// save as prior width
					priorWidth = match - i;
				
					// move ahead
					i = match - 1;
				}
				
			}
			
		}
		
		// mark all frequencies of zero as silence
		for (int i = 0; i < this.samples.length; i++) {
			if (this.samples[i] == 0) {
				this.samples[i] = -1;
			}
		}
		
		// convert frequency to pitch numbers
		for (int i = 0; i < this.samples.length; i++) {
			if (this.frequency[i] > 0) {
				this.midiValue[i] = frequencyToMidi( this.frequency[i] );
			} else {
				this.midiValue[i] = -1;
			}			
		}
		
	}
	
	
	public void findSegments() {
		// clear the segments
		this.segments = new Vector();
		
		// start segment not set
		int startAt = -1;
		double midi = 0;
		double midiSum = 0;
		
		// iterate through the 
		for (int i = 0; i < this.samples.length; i++) {
			// found silence?
			if (this.frequency[i] == -1 || 
				(startAt != -1 && Math.abs(this.midiValue[i]-midi) > 2.0 )) {
				 // got a start segment?
				if (startAt != -1) {
					// how wide is the segment?
					// FIXME: need real value at some point, not 1000
					if ( i - startAt > 1000) {
						int width = i - startAt;
						int averagedMidi = (int)((midiSum / width) + 0.5);
						WaveSegment segment = new WaveSegment(startAt, i-1, averagedMidi, this.samples);
						this.segments.add(segment);
					}
				}
				// clear the start
				startAt = -1;
			} else {
				// need to set start position?
				if (startAt == -1) {
					// set start
					startAt = i;
					// FIXME: need to find median value or something...
					midi = this.midiValue[i];
					midiSum = this.midiValue[i];
				} else {
					// accumulate midi value
					midiSum += this.midiValue[i];
				}
			}
		}
		
	}
	
	public static double frequencyToMidi( float f ) {
		return ((12.0 * Math.log(f/220.0)) / Math.log(2.0)) + 57.01d; 
	}
	
	public static double midiToFrequency( double midi ) {
		return 440.d * Math.pow(2, (midi-69)/12 );
	}
	
}
