//WaveBuilder.java
//Resample a wave to the target frequency
// This is a static helper class for the Analysis class
//
//(c) 2005 David Cuny
//http://tunafish.sourceforge.net
//Released under Qt Public License

public class WaveBuilder {

	public static final int USE_CATMULL_CLARK = 0;
	public static final int USE_CUBIC_SPLINE = 1;
	public static final int USE_PSOLA = 2;
	
	public static int interpolation = USE_CATMULL_CLARK; 
	
	/**
	 * Given four equally spaced sample points p0, p1, p2 and p3
	 * and a value t ranging from 0 (t=p1) and 1 (t=p2), return
	 * an interpolated value between p1 and p2.
	 * @param p0 First of four equally spaces samples
	 * @param p1 Second of four equally spaces samples
	 * @param p2 Third of four equally spaces samples
	 * @param p3 Fourth of four equally spaces samples
	 * @param t Position between p1 and p2, ranging from 0 (p1) and 1 (p2)
	 * @return interpolated value between p1 and p2
	 */
	final static double catmullClarkSpline( double p0, double p1, double p2, double p3, float t ) {
		// calculate t squared and t cubed
		double t2 = t*t;
		double t3 = t2 * 2;
		
		// calculate spline value
		return 0.5d * ( 
			( 2d*p1 ) + 
			(-p0 + p2) * t + 
			(2d*p0 - 5.0*p1 + 4*p2 - p3) * t2 +
			(-p0 + 3d*p1 - 3d*p2 + p3) * t3
		);
	}

	/**
	 * Given four equally spaced sample points p0, p1, p2 and p3
	 * and a value t ranging from 0 (t=p1) and 1 (t=p2), return
	 * an interpolated value between p1 and p2.
	 * @param p0 First of four equally spaces samples
	 * @param p1 Second of four equally spaces samples
	 * @param p2 Third of four equally spaces samples
	 * @param p3 Fourth of four equally spaces samples
	 * @param t Position between p1 and p2, ranging from 0 (p1) and 1 (p2)
	 * @return interpolated value between p1 and p2
	 */
	final static double cubicSpline( double p0, double p1, double p2, double p3, float t ) {
		// calculate the weights
		double w0 = p1;
		double w1 = (-2d*p0 - 3d*p1 + 6d*p2 - p3) / 6d;
		double w2 = (3d*p0 - 6d*p1 + 3d*p2) / 6d;
		double w3 = (-p0 + 3d*p1 - 3d*p2 + p3) / 6d;
		
		// calcuate squared and cubed values of t
		double t2 = t * t;
		double t3 = t2 * t;
		
		// return interpolated value
		return w0 + w1*t + w2*t2 + w3*t3;
	}
	
	
	final static int getInterpolatedValue( float index, int samples[] ) {
		
		// check range
		if (index < 4 || index > (samples.length - 4)) {
			// not enough points to interpolate
			return samples[(int)index];
		}
		
		// get integer portion
		int i = (int)index;
		
		// get fractional portion
		float t = index - i;
		
		// get sample points
		double p0 = (double)samples[i-1];
		double p1 = (double)samples[i];
		double p2 = (double)samples[i+1];
		double p3 = (double)samples[i+2];
	
		// which interpolation method?
		if (interpolation == USE_CATMULL_CLARK) {
			// catmull clark
			return (int)catmullClarkSpline(p0, p1, p2, p3, t );
			
		} else if (interpolation == USE_CUBIC_SPLINE) {
			// cubic spine
			return (int)cubicSpline(p0, p1, p2, p3, t );
		} else {
			// psola
			return samples[i];
		}
	}

	final static int interpolateSingleWave( int srcSamples[], int srcAt, int width, int dstSamples[], int dstAt, double rate ) {

		// position to take sample from
		float sampleAt = (float)srcAt;
		
		
		// end location of the wave
		int stopAt = srcAt + width;
		
		// iterate over the wave
		while (sampleAt < stopAt && dstAt < dstSamples.length) {
			// interpolate the value
			dstSamples[dstAt] = getInterpolatedValue(sampleAt, srcSamples);
			// move ahead
			dstAt++;
			sampleAt += rate;
		}
		
		// return next position to write to
		return dstAt;
	}
	
	final static int[] resampleStream( int samples[], float frequency[], double targetMidi[], int waveWidth[] ) {
		// create a buffer to hold the new samples
		int buffer[] = new int[samples.length];

		int src = 0;
		int dst = 0;
		
		// iterate over the old wave
		while (src < samples.length && dst < samples.length) {
			
			// not the start of a wave?
			if (waveWidth[src] == 0) {
				// copy verbatim
				buffer[dst++] = samples[src++];
				
				// jump back to the top of the loop
				continue;
			}
						
			
System.out.println("found wave start at " + src);
			
			// get the frequency
			double actualFrequency = frequency[src];
			
			// get the target frequency
			double targetFrequency = Analysis.midiToFrequency( targetMidi[src] );
			
			// calculate the amount to scale the wave
			double rate = targetFrequency / actualFrequency;

			// get the width of the wave
			int width = waveWidth[src];
				
			// this shouldn't happen...
			if (width < 1) {
				System.out.println("zero length found at " + src);
				System.exit(0);
			}
			
			// width of destination
			int dstWidth = (int)(width/rate);

			// this shouldn't happen...
			if (dstWidth < 1) {
				System.out.println("zero destination width - " + width + " divided by rate of " + rate );
				System.exit(0);
			}

			// keep the output in sync
			while (dst+dstWidth < src+width && dstWidth > 0) {
				// what resampling method?
				if (WaveBuilder.interpolation == USE_PSOLA) {
					// copy wave to destination, adding to prior value
					for (int i = 0; i < width*2; i++) {
						// exit if past end
						if (dst+i >= samples.length) break;

						
						// get a sample
						int sample = 0;
						if (src + i < samples.length) {
							sample = samples[src+i];
						}
						
						// average sample
						// FIXME: this is a hack
						buffer[dst+i] = (sample + buffer[dst+i]) / 2;
						
					}
						
					// move destination ahead
					dst += dstWidth;

				} else {
					// interpolate a new wave
					dst = interpolateSingleWave( samples, src, width, buffer, dst, rate);
				}
			}
				
			// move source ahead
			src += width;				
		}
		
		// return the new wave
		return buffer;
		
	}
	
	// FIXME: this should be a single routine
	public final static int bytesToSample( byte buffer[], int index, int bytesPerSample, int bitsPerSample ) {
		// return the value of a sample at byte position

		int result = 0; 

		// scale index by width of samples
		index *= bytesPerSample;
		
		if (index >= buffer.length ) {
			// System.out.println("index " + index + " out of range " + this.byteCount);
			return 0;
		}

		// get the low byte, convert from two's complement to unsigned
		int lsb = (int)buffer[index];
		
		switch (bytesPerSample) {
		case 1:
			// 8 bit unsigned
			return signedToUnsigned( lsb ); 			
			
		case 2:			
			// shouldn't be signed			
			if (lsb < 0) lsb = -lsb;
			int msb = (int)buffer[index+1];						
					
			// combine						
			result = lsb + (msb * 256);
								
			// remove unneeded bits?
			int padBits = 16 - bitsPerSample;
			if (padBits != 0) {
				result = result >> (padBits);
			}
					
		}
		return result;
	}

		
	public final static int signedToUnsigned( int n ) {
		if (n < 0) {
			// negate
			n = -n;
			
			// complement
			n++;
			
			// invert
			n = ~n;
			
			// mask off high bits
			n &= 0xFF;			
		}
		return n;
	}

	public final static int unsignedToSigned( int n ) {
		// high bit set?
		if ((n & 128) != 0) {
			// subtract one
			n--;
			// complement
			n = ~n;
			// take low bytes
			n = 0xFF & n;
			// make negative
			n = -n;
		}
		return n;
	}
	
	
	public final static void samplesToBytes( int samples[], byte bytes[], int bytesPerSample ) {
		
		// build buffer from samples

		int bufferAt = 0; 
		for (int i = 0; i < samples.length; i++) {
			
			// get a sample
			int value = samples[i];
			
			switch (bytesPerSample) {
			case 1:
				// 8 bit unsigned
				bytes[bufferAt++] = (byte)unsignedToSigned(value);
				break; 			
			
			case 2:
				byte lsb = (byte)(0xFF & value);
				byte msb = (byte)((0xFF00 & value) >> 8);
								
				// FIXME: handle padBits
			
				bytes[bufferAt++] = lsb; 
				bytes[bufferAt++] = msb;
				break;
				
			default:
				Utilities.errMessage("Can't handle wave with " + bytesPerSample + " bytes per sample." );
				System.exit(0);
						
			}			
		}		
	}	
	
}

