//  WaveFile.java
// Load, display and play wave files
//
// (c) 2005 David Cuny
// http://tunafish.sourceforge.net
// Released under Qt Public License

import java.awt.Color;
import java.awt.Graphics;
import java.io.ByteArrayInputStream;
import java.io.File;
import java.util.Vector;

import javax.sound.sampled.AudioFormat;
import javax.sound.sampled.AudioFileFormat;
import javax.sound.sampled.AudioInputStream;
import javax.sound.sampled.AudioSystem;
import javax.sound.sampled.DataLine;
import javax.sound.sampled.FloatControl;
import javax.sound.sampled.SourceDataLine;
import javax.sound.sampled.AudioFormat.Encoding;
import javax.swing.JFileChooser;
import javax.swing.JOptionPane;

public class WaveHandler {
	
	static final int BAR_WIDTH = 30;
	static final int WAVE_HEIGHT = 60;
	static int SEGMENT_HEIGHT = 12;

	// zoom factor
	float zoom = 512;
	
	// wave correction rate
	static double CORRECTION_RATE = .0003;
	
	// buffers holding the wave data as samples and bytes 
	byte[] rawBytes = null;	
	int[] originalSamples = null;
	int[] modifiedSamples = null;
		
	float sampleRate;
	int byteCount = 0;
	int sampleCount = 0;
	int bitsPerSample = 0; // number of signifcant bits
	int bytesPerSample = 1; // 1 = 8 bit, 2 = 16 bit
	int channels = 1; // mono or stereo
	boolean signed = false;	
	boolean needsResampling = false;
	
	// playback volume
	float playBackVolume = 5;
	
	// selection
	int selectionStart = -1;
	int selectionEnd = -1;
	

	// allow stopping of wave playback
	static boolean stopPlayingWave = false;
		
	// name of file
	String fileName = new String("");
	File file = null;

	// output data line
	SourceDataLine sourceDataLine = null;
	
	// audio format
	AudioFormat sourceAudioFormat = null;
	
	// stream
	AudioInputStream sourceAudioInputStream = null;

	Analysis analyzeWave;
	
	public void fileOpen() {

		// autosave?
		if (cancelledAutoSave()) {
			return;
		}

		// open file dialog
		JFileChooser fc = Utilities.fileChooser(null);
		fc.addChoosableFileFilter(new WavFilter());		
		int returnVal = Utilities.showOpenDialog(fc);
		if (returnVal == JFileChooser.APPROVE_OPTION) {
			// attempt to load the file
			this.loadWave( fc.getSelectedFile().getPath() );
		}
	}	

	public void fileSave() {
		if (this.file == null) {
			// use save as instead
			fileSaveAs();
		} else {
			// write out the wave
			this.writeFile( this.file.getPath() );
		}
	}


	public void fileSaveAs() {
		JFileChooser fc = Utilities.fileChooser(this.file);
		fc.addChoosableFileFilter(new WavFilter());
		int returnVal = Utilities.showSaveDialog(fc);
		if (returnVal == JFileChooser.APPROVE_OPTION) {
			// get the selected file
			File file = fc.getSelectedFile();

			// make sure it ends with the proper extention
			String fileName = file.getPath();
			fileName = addFileExtention(fileName, WavFilter.EXTENTION);

			// write the wave file out
			writeFile(fileName);
			
			// FIXME: handle errors
			// make this the current file
			this.fileName = fileName;
			this.file = file;
		}
	}

	// ensure file name ends with given extention
	public String addFileExtention(String fileName, String extention) {

		String lowerFileName = fileName.toLowerCase();
		if (!(lowerFileName.endsWith(extention))) {
			return fileName + extention;
		}
		return fileName;
	}

	
	// if returns false, user cancelled autosave action
	public boolean cancelledAutoSave() {
		// no work to save?
		if (this.file == null) {
			return false;
		}

		// ask user if they want to save their current work
		int result = JOptionPane.showConfirmDialog(null,
				"Save Current File First?", "Save File?",
				JOptionPane.YES_NO_CANCEL_OPTION);
		if (result == JOptionPane.CANCEL_OPTION) {
			// cancelled
			return true;

		} else if (result == JOptionPane.NO_OPTION) {
			// didn't save, but didn't cancel
			return false;
		}

		// make sure it ends with the proper extention
		String openFileName = this.file.getPath();
		openFileName = addFileExtention(openFileName, ".wav");

		// write the new wave file out
		writeFile( openFileName ); 

		return false;

	}

	public void startSelection( int at ) {
		System.out.println("startSelection at " + at);
		this.selectionStart = at;
		this.selectionEnd = -1;
	}
	
	public void extendSelection( int at ) {
		if (at < this.selectionStart) {
			System.out.println("extendSelection at " + at);
			this.selectionEnd = this.selectionStart;
			this.selectionStart = at;
		} else {
			System.out.println("extendSelection end at " + at);
			this.selectionEnd = at;
		}
	}

	public String midiNoteName( int i ) {
		int octave = (i / 12) - 2;
		switch (i % 12) {
		case 0: return "C" + octave;
		case 1: return "C#" + octave;
		case 2: return "D" + octave;
		case 3: return "D#" + octave;
		case 4: return "E" + octave;
		case 5: return "F" + octave;
		case 6: return "F#" + octave;
		case 7: return "G" + octave;
		case 8: return "G#" + octave; 
		case 9: return "A" + octave;
		case 10: return "A#" + octave;
		case 11: return "B" + octave;
		}
		return "?";		
	}

	// convert midi pitch to screen position
	public int midiToPos( double midi, int height ) {
		return height - (int)((midi-40)*SEGMENT_HEIGHT);

	}
	
	// FIXME: needs to add x offset
	public WaveSegment hitSegment( int x, int y, int height ) {
		// no waveform?
		if (byteCount == 0) {
			return null;
		}

		// get the wave segments
		Vector segments = this.analyzeWave.segments;
		
		// iterate over them
		for (int i = 0; i < segments.size(); i++) {
			// get a segment
			WaveSegment segment = (WaveSegment)segments.get(i);
			
			// calculate the position on the screen
			int xStart = (int)(segment.startSample / this.zoom);
			int yStart = midiToPos( segment.midiNumber, height );
			
			// render the segment
			segment.calcScreenPosition( xStart, yStart, this.zoom );
		
			// inside segment?
			if (segment.x1 < x && segment.x2 > x && segment.y1 < y && segment.y2 > y) {
				// return hit segment
				return segment;
			}
		}

		// no segement hit
		return null;
		
	}
	
	
	public boolean isBlackKey( int midiNote ) {
		int scaleNumber = midiNote % 12;
		switch (scaleNumber) {
			case 1:
			case 3:
			case 6:
			case 8:
			case 10:
			case 11:
				// black key
				return true;
			default:
				// white key
				return false;
		}
	}
	
	
	public void render(Graphics g, int height, int width, int waveOffset) {

				
		// draw the lines for the midi pitch scale
		for (int i = 88; i > 20; i--) {
			// FIXME: make sure this is in range, to speed rendering
			
			// FIXME: eventually change to reflect scale notes
			if (isBlackKey(i)) {
				g.setColor(ColorScheme.SCALE_BACKGROUND);
			} else {
				g.setColor(ColorScheme.NON_SCALE_BACKGROUND);
			}
			int y = this.midiToPos( i-.5, height );
			g.fillRect( 0, y, width, y+SEGMENT_HEIGHT );
			
			// draw a dividing line
			g.setColor( ColorScheme.SCALE_DIVIDING_LINE);
			g.drawLine( 0, y, width, y );
			
			// draw the note name
			g.setColor(ColorScheme.PITCH_NAME_FOREGROUND);
			y = this.midiToPos( i-.5, height );
			g.drawString( midiNoteName(i), 5, y );
		}
		
		
		// is there a selection?
		if (selectionStart != -1) {
			g.setColor(ColorScheme.SELECTION);
			int start = (int)(selectionStart / zoom) - waveOffset;
			if (selectionEnd == -1) {
				// only draw selection line
				g.fillRect( start-waveOffset, 0, 1, height );
			} else {				
				int end = (int)(selectionEnd / zoom) - waveOffset;
				// draw selection area
				g.fillRect( start, 0, end-start, height );			

			}
		}

		
		// no waveform?
		if (byteCount == 0) {
			return;
		}
		
		// HACK!!!
		// is there a selection?
		if (selectionStart != -1) {
			int sampleCount = 1024;
			
			// select a window of 512 samples
			float[] sourceSamples = new float[sampleCount];
			
			// copy the samples into the array
			for (int i = 0; i < sampleCount; i++) {
				sourceSamples[i] = this.originalSamples[selectionStart+i];
			}
			
			// analyze them
			float mag[] = FastFourierTransform.fftMag(sourceSamples);
			
			// result is half the size of the sample count
			sampleCount /= 2;
			
			// build a polyline
			int maxAt = 0;
			int magX[] = new int[sampleCount];
			int magY[] = new int[sampleCount];
			for (int i = 0; i < sampleCount; i++) {
				// copy to polyline
				magX[i] = i + 20;
				magY[i] = (int)(mag[i]*0.05);
				
				// largest so far?
				if (mag[maxAt] < mag[i]) {
					// save value
					maxAt = i;
				}
			}
			
			// render the plot
			g.setColor(Color.orange);
			g.drawPolyline( magX, magY, sampleCount );
			
			// calculate frequency
		    float f = 0.5f * ((float)this.sampleRate * (float)maxAt) / (float)sampleCount;
			
			// display the value
			g.setColor(Color.black);
			g.drawString( "Frequency=" + f + "Hz", 10, 10);

			
		}
				
		// get the wave segments
		Vector segments = this.analyzeWave.segments;
		
		// iterate over them
		for (int i = 0; i < segments.size(); i++) {
			// get a segment
			WaveSegment segment = (WaveSegment)segments.get(i);
			
			// calculate start position on the screen
			int xStart = (int)(segment.startSample / zoom);
			int yStart = midiToPos( segment.midiNumber, height );
			
			// render the segment
			segment.render(g, xStart, yStart, zoom, this.modifiedSamples);			
		}

		

		g.setColor(Color.black);
		g.drawString( "File = ", 10, height-10);
		//g.drawString( "Zoom = " + zoom, 10, height-30);
		if (selectionStart != -1) {
			//g.drawString( "Point " + selectionEnd + "=" + this.analyzeWave.samples[selectionEnd], 10, 30);
		}
		
		//System.out.println("bits per sample =" + this.bitsPerSample);
		//System.out.println("bytes per sample =" + this.bytesPerSample);
		//System.out.println("channels =" + this.channels);
		//System.out.println("signed =" + this.signed);
				
		// center of wave
		int center = height - WAVE_HEIGHT/2;

		// line drawing is really, really slow, so we cache the points and use drawPolyline
		int xPos[] = new int[width];
		int yPos[] = new int[width];

		// cache points for frequency and corrected frequency
		int xFreqPos[] = new int[width];
		int yFreqPos[] = new int[width];		
		int yFixedFreqPos[] = new int[width];		
		int freqWidth = 0;

		// cache points for corrected frequency
		int newFreqPos[] = new int[width];
		
		// draw the wave
		g.setColor(ColorScheme.WAVE_COLOR);
		float at = waveOffset;
		int zero = this.modifiedSamples[0];
		int halfHeight = WAVE_HEIGHT / 2;
		for (int x = 0; x < width; x++) {
			
			// increment position
			at += zoom;
			int i = (int)at;
			
			// past range?
			if (i >= this.sampleCount) {
				break;
			}

			// get sample value for frame
			int value = 0;
			
			// invert the value for display
			value = -this.modifiedSamples[i] - zero;
			 
			// fit sample to the window
			value = center + (int)(halfHeight *((float)value/(float)this.analyzeWave.maxAmplitude));			
						
			// store for polyline
			xPos[x] = x;
			yPos[x] = value;

			// store frequency
			if (i < this.sampleCount) {
				// get analyzed pitch
				double midi = this.analyzeWave.midiValue[i];
				yFreqPos[freqWidth] = midiToPos( midi, height ); 
				
				// get corrected pitch
				midi = this.analyzeWave.correctedMidiValue[i];
				yFixedFreqPos[freqWidth] = midiToPos( midi, height );
				
				// x position
				xFreqPos[freqWidth] = x;
				
				freqWidth += 1;
				
			}
			
			// draw the accumulated wave?
			// FIXME: bottom test is a hack...
			if (this.analyzeWave.frequency[i] < 20 || x == width-1) {
				// draw the frequency?
				if (freqWidth > 0) {
					
					// ignore this point
					freqWidth -= 1;
					
					// frequency
					g.setColor(Color.red);
					g.drawPolyline( xFreqPos, yFreqPos, freqWidth );
			
					// corrected frequency
					g.setColor(Color.green);
					g.drawPolyline( xFreqPos, yFixedFreqPos, freqWidth );
					
					// reset the wave
					freqWidth = 0;
				}
			}



			
		}
		
		// draw the wave
		g.setColor(ColorScheme.OUTER_LINE);
		g.drawPolyline( xPos, yPos, width );				
				
	}

				
	private final File askUserForFile(String fileName) {
		
		// prompt user
		int result = JOptionPane.showConfirmDialog(null,
				"Can't find the file '" + fileName
						+ "'\nDo you want to find it manually?",
				"Unable to find .wav file", JOptionPane.YES_NO_OPTION);

		if (result == JOptionPane.YES_OPTION) {					
			// open file dialog
			JFileChooser fc = Utilities.fileChooser(null);

			// limit to .wav files
			fc.addChoosableFileFilter(new WavFilter());
			
			// open the dialog			
			int returnVal = Utilities.showOpenDialog(fc);
			
			// check results
			if (returnVal == JFileChooser.APPROVE_OPTION) {
				return fc.getSelectedFile();
			}
		}
		
		// no file
		return null;

	}

	
	// write a waveform to a file
	public final void writeFile(String fileName) {
		
		try {
		
			// create a file to write to
			File outFile = new File( fileName );
		
			// create an input stream with the data
			ByteArrayInputStream inputStream = new ByteArrayInputStream(this.rawBytes);
		
			long frameSize = this.byteCount / sourceAudioFormat.getFrameSize();
		
			// convert it to an audio stream		
			AudioInputStream outputStream = new AudioInputStream(
				inputStream, sourceAudioFormat, frameSize );
				
			// write out to a file
			int writtenBytes = AudioSystem.write(outputStream, AudioFileFormat.Type.WAVE, outFile);
			
		} catch (Exception e) {
			Utilities.errMessage( "Error writing file '" + fileName + "'" );
		}
	}
	
	final void readWaveStream( File file ) throws Exception {
		
		// try to open the input stream
		this.sourceAudioInputStream = null;
		this.sourceAudioInputStream = AudioSystem.getAudioInputStream( file );
		
		// get the audio format
		this.sourceAudioFormat = sourceAudioInputStream.getFormat();
						
		// get the input stream
		AudioInputStream audioInputStream = AudioSystem.getAudioInputStream(sourceAudioFormat, sourceAudioInputStream );
		
		// number of channels
		channels = sourceAudioFormat.getChannels();

		// bits per sample
		bitsPerSample = sourceAudioFormat.getSampleSizeInBits();
		
		// bytes per sample
		bytesPerSample = (channels * sourceAudioFormat.getSampleSizeInBits())/ 8;
		
		// get the sample rate
		this.sampleRate = sourceAudioFormat.getSampleRate();		
		
		// signed?
		signed = sourceAudioFormat.getEncoding().equals(Encoding.PCM_SIGNED);
		
		// calculate the number of bytes in the stream
		byteCount = (int)(audioInputStream.getFrameLength() * sourceAudioFormat.getFrameSize());
		
		// calcuate the metrics of the wave
		sampleCount = byteCount / (bytesPerSample * channels);
		        
		// allocate and load the array
		rawBytes = new byte[byteCount];
	    audioInputStream.read(rawBytes, 0, byteCount);
	    
		// allocate room for the samples
		this.originalSamples = new int[sampleCount];
		// FIXME: hack
		this.modifiedSamples = new int[sampleCount];
		for (int i = 0; i < sampleCount; i++) {
			originalSamples[i] = WaveBuilder.bytesToSample(rawBytes, i, this.bytesPerSample, this.bitsPerSample);
			// FIXME: hack
			modifiedSamples[i] = originalSamples[i];
		}	    
	}
	

	// load a waveform from a file
	public final void loadWave(String fileName) {
				
		// FIXME! If the bytes per sample changes, need to open a new dataline!
		
		// try loading the wave file
		try {
			// open the file
			this.file = new File(fileName);
			if (!file.canRead()) {
				// throw an exception
				throw new Exception();
			}
			
			// set the filename
			this.fileName = file.getPath();
						
		} catch (Exception e) {
			// can't find or open the file; ask user for help
			file = askUserForFile(fileName);
			if (file == null) {
				// clear flags and exit
				byteCount = 0;
				this.file = null;
				this.fileName = "";
				return;
			}
		} 

		// read the data from the file
		try {
			readWaveStream(this.file);
		} catch (Exception e) {
			Utilities.errMessage( "Error reading file '" + file.getName() + "'" );
			byteCount = 0;
			return;
		}
		
		// save the name
		fileName = file.getPath();
		
		// set the title
		Main.mainWindow.setTitle(file.getName());

		// need to close prior line?
		if (sourceDataLine != null) {
			sourceDataLine.flush();
			sourceDataLine.close();
			sourceDataLine = null;
		}
		
		// need to open a line?
		if (sourceDataLine == null) {
			// the data line
			DataLine.Info info = new DataLine.Info(SourceDataLine.class, sourceAudioFormat);
			try {				
				// open the line
				sourceDataLine = (SourceDataLine) AudioSystem.getLine(info);
				sourceDataLine.open(sourceAudioFormat);
				
				// prepare the line for output
				sourceDataLine.start();

			} catch (Exception e) {
				Utilities.errMessage("Unable to open audio device");
				sourceDataLine = null;
			}
		} else {
			// is the line open?
			if (!(sourceDataLine.isOpen())) {
				Utilities.errMessage("Data line is not open");
				sourceDataLine = null;
			}

		}
		
		// adjust the scrollbar
		Main.mainWindow.wavePanel.adjustScrollbar();
				
		// analyze the wave
		this.analyzeWave = new Analysis(this.originalSamples, this.sampleRate, 1024);
		
		// find the frequencies
		this.analyzeWave.findFrequencies();
		
		// find the individual segments
		this.analyzeWave.findSegments();
		
		// no pitch correction
		this.restoreOriginal(); 
		
	}


	public void restoreOriginal() {
		// set the scale
		this.analyzeWave.setScale( "Chromatic" );

		// copy original samples
		for (int i = 0; i < this.originalSamples.length; i++) {
			this.modifiedSamples[i] = this.originalSamples[i];
		}
		
		// set correction rate to zero
		double oldRate = this.CORRECTION_RATE;
		this.CORRECTION_RATE = 0.0;
		this.analyzeWave.calculateCorrectPitch();
		
		// restore correction rate
		this.CORRECTION_RATE = oldRate;
		
	}
	
	public void resampleWaveIfRequired() {
		
		// skip if no analysis
		if (this.analyzeWave == null) return;
		
		// needs resampling?
		if (this.analyzeWave.needsResampling) {		
			
			// resample the wave to the new scale
			this.modifiedSamples = WaveBuilder.resampleStream( this.originalSamples, this.analyzeWave.frequency, this.analyzeWave.correctedMidiValue, this.analyzeWave.waveWidth );
		
			// convert samples back to bytes for playback
			WaveBuilder.samplesToBytes( this.modifiedSamples, this.rawBytes, this.bytesPerSample );
			
			// clear flag
			this.analyzeWave.needsResampling = false;

		}
	}
	
	public void setVolume( float newVolume ) {
		// this handles events from the volume slider control
		
		// store the value in playbackVolume
		this.playBackVolume = newVolume;
		
		// adjust the volume if a dataline is open
		adjustVolume();
	}
		
	
	public void adjustVolume() {
		// is there an open dataline?
		if (this.sourceDataLine != null) {
			// get the volume control from the dataline			
			FloatControl volumeControl = (FloatControl) this.sourceDataLine.getControl( FloatControl.Type.MASTER_GAIN );
			
			// scale the playback volume to a range the dataline expects
			float maxVolume = volumeControl.getMaximum();
			float minVolume = volumeControl.getMinimum();
			float volume = minVolume + ((this.playBackVolume/10f)*(maxVolume - minVolume));
			volumeControl.setValue( volume );
		}		 
	}

	// play the wave file
	public final void playWave(int startSample, int endSample ) {

		// resample the wave, if needed
		this.resampleWaveIfRequired();

		
		// no start sample?
		if (startSample == -1) {
			startSample = 0;
		}

		// no end frame or past end
		if (endSample == -1 || endSample > originalSamples.length ) {
			endSample = originalSamples.length;
		}
 
		// play the thread on a background thread
		try {
			// calculate start and end byte sample
			int startByte = startSample * this.bytesPerSample;
			int endByte = endSample * this.bytesPerSample;
			
			// start the thread
			new PlayThread(startByte, endByte, rawBytes).start();
		} catch (Exception e) {
			Utilities.errMessage("Error playing the wave file");
		}
	}

	// play the sound on a background thread
	class PlayThread extends Thread {

		int startByte, endByte;
		byte[] samples;

		public PlayThread(int startAtByte, int endAtByte, byte samples[]) {
			this.startByte = startAtByte;
			this.endByte = endAtByte;
			this.samples = samples;
			stopPlayingWave = false;
		}

		public final void run() {

			// make sure a waveform is loaded
			if (byteCount == 0) {
				return;
			}

			// make sure audio device connected
			if (sourceDataLine == null) {
				Utilities.errMessage("Audio device failed. Try reloading file");
				return;
			}

			// is it open?
			if (!(sourceDataLine.isOpen())) {
				Utilities.errMessage("Data line is not open");
				sourceDataLine = null;
			}
			
			// adjust the dataline volume
			adjustVolume();
			
			// send out 1K at a time
			int bufferSize = (1024 * 4);						

			// calculate buffer
			int startOfBuffer = startByte;

			while (!WaveHandler.stopPlayingWave) {
				// check for buffer overflow
				if ((startOfBuffer + bufferSize) >= endByte) {
					bufferSize = (endByte) - startOfBuffer;
				}

				// start the line
				try {
					sourceDataLine.start();
				} catch (Exception e) {
					// oops
					System.out.println("Unable to start the line");
				}
				
				// fixme: throws illegal length if too large				
				int nBytesWritten = sourceDataLine.write(samples, startOfBuffer, bufferSize);

				// point to next starting point in the buffer
				startOfBuffer += bufferSize;
				if (startOfBuffer >= endByte) {
					break;
				}

			}
			
			// play out the remainder of the line, but don't close it
			sourceDataLine.stop();
			sourceDataLine.drain();
			sourceDataLine.flush();

		}

		
	}
	
}
